@@ -1,4 +1,4 @@
-from __future__ import print_function, unicode_literals
+from __future__ import absolute_import, division, print_function, unicode_literals
 
 import sys
 import collections
@@ -7,10 +7,15 @@ import re
 import functools
 from unittest.case import TestCase
 from collections import namedtuple, Mapping, Iterable
+from six import string_types
+
+from ._types import NumericString, NumericByteString, IntegerString, IntegerByteString
+
 
 USING_PYTHON2 = True if sys.version_info < (3, 0) else False
 
-__all__ = ['EnsureError', 'Ensure', 'Check', 'ensure', 'check', 'ensure_raises', 'ensure_raises_regex', 'ensure_annotations']
+__all__ = ['EnsureError', 'Ensure', 'Check', 'ensure', 'check', 'ensure_raises', 'ensure_raises_regex',
+           'ensure_annotations']
 
 if USING_PYTHON2:
     __all__ = map(bytes, __all__)
@@ -102,12 +107,20 @@ class MappingInspector(Inspector):
         return self.to(prototype)
 
 
-class AttributeInspector(Inspector):
+class ReferenceInspector(Inspector):
     @property
     def which(self):
         return Ensure(self._subject)
 
 
+class ChainInspector(Inspector):
+    @property
+    def also(self):
+        return Ensure(self._subject)
+
+    def __repr__(self):
+        return "{}: OK".format(Inspector.__repr__(self))
+
 class KeyInspector(Inspector):
     @property
     def whose_value(self):
@@ -192,12 +205,14 @@ class Ensure(Inspector):
         Ensures :attr:`subject` is equal to *other*.
         """
         self._run(unittest_case.assertEqual, (self._subject, other))
+        return ChainInspector(self._subject)
 
     def is_not_equal_to(self, other):
         """
         Ensures :attr:`subject` is not equal to *other*.
         """
         self._run(unittest_case.assertNotEqual, (self._subject, other))
+        return ChainInspector(self._subject)
 
     does_not_equal = is_not_equal_to
 
@@ -206,18 +221,21 @@ class Ensure(Inspector):
         Ensures :attr:`subject` is *other* (object identity check).
         """
         self._run(unittest_case.assertIs, (self._subject, other))
+        return ChainInspector(self._subject)
 
     def is_not(self, other):
         """
         Ensures :attr:`subject` is not *other* (object identity check).
         """
         self._run(unittest_case.assertIsNot, (self._subject, other))
+        return ChainInspector(self._subject)
 
     def contains(self, element):
         """
         Ensures :attr:`subject` contains *other*.
         """
         self._run(unittest_case.assertIn, (element, self._subject))
+        return ChainInspector(self._subject)
 
     def contains_none_of(self, elements):
         """
@@ -225,6 +243,7 @@ class Ensure(Inspector):
         """
         for element in elements:
             self._run(unittest_case.assertNotIn, (element, self._subject))
+        return ChainInspector(self._subject)
 
     def contains_one_of(self, elements):
         """
@@ -232,6 +251,7 @@ class Ensure(Inspector):
         """
         if sum(e in self._subject for e in elements) != 1:
             raise self._error_factory(_format("Expected {} to have exactly one of {}", self._subject, elements))
+        return ChainInspector(self._subject)
 
     def contains_only(self, elements):
         """
@@ -240,8 +260,9 @@ class Ensure(Inspector):
         for element in self._subject:
             if element not in elements:
                 raise self._error_factory(_format("Expected {} to have only {}, but it contains {}",
-                                                 self._subject, elements, element))
+                                                  self._subject, elements, element))
         self.contains_all_of(elements)
+        return ChainInspector(self._subject)
 
     def contains_some_of(self, elements):
         """
@@ -249,6 +270,7 @@ class Ensure(Inspector):
         """
         if all(e not in self._subject for e in elements):
             raise self._error_factory(_format("Expected {} to have some of {}", self._subject, elements))
+        return ChainInspector(self._subject)
 
     contains_one_or_more_of = contains_some_of
 
@@ -259,13 +281,15 @@ class Ensure(Inspector):
         for element in elements:
             if element not in self._subject:
                 raise self._error_factory(_format("Expected {} to have all of {}, but it does not contain {}",
-                                                 self._subject, elements, element))
+                                                  self._subject, elements, element))
+        return ChainInspector(self._subject)
 
     def does_not_contain(self, element):
         """
         Ensures :attr:`subject` does not contain *element*.
         """
         self._run(unittest_case.assertNotIn, (element, self._subject))
+        return ChainInspector(self._subject)
 
     def contains_no(self, prototype):
         """
@@ -273,6 +297,7 @@ class Ensure(Inspector):
         """
         for element in self._subject:
             self._run(unittest_case.assertNotIsInstance, (element, prototype))
+        return ChainInspector(self._subject)
 
     def has_key(self, key):
         """
@@ -288,6 +313,7 @@ class Ensure(Inspector):
         """
         self.is_a(Mapping)
         self.contains_all_of(keys)
+        return ChainInspector(self._subject)
 
     def has_only_keys(self, keys):
         """
@@ -295,6 +321,7 @@ class Ensure(Inspector):
         """
         self.is_a(Mapping)
         self.contains_only(keys)
+        return ChainInspector(self._subject)
 
     def has_attribute(self, attr):
         """
@@ -302,7 +329,7 @@ class Ensure(Inspector):
         """
         if not hasattr(self._subject, attr):
             raise self._error_factory(_format("Expected {} to have attribute {}", self._subject, attr))
-        return AttributeInspector(getattr(self._subject, attr))
+        return ReferenceInspector(getattr(self._subject, attr))
 
     hasattr = has_attribute
 
@@ -311,12 +338,14 @@ class Ensure(Inspector):
         Ensures :attr:`subject` is contained in *iterable*.
         """
         self._run(unittest_case.assertIn, (self._subject, iterable))
+        return ChainInspector(self._subject)
 
     def is_not_in(self, iterable):
         """
         Ensures :attr:`subject` is not contained in *iterable*.
         """
         self._run(unittest_case.assertNotIn, (self._subject, iterable))
+        return ChainInspector(self._subject)
 
     not_in = is_not_in
 
@@ -325,24 +354,40 @@ class Ensure(Inspector):
         Ensures :attr:`subject` is ``True``.
         """
         self._run(unittest_case.assertTrue, (self._subject,))
+        return ChainInspector(self._subject)
 
     def is_false(self):
         """
         Ensures :attr:`subject` is ``False``.
         """
         self._run(unittest_case.assertFalse, (self._subject,))
+        return ChainInspector(self._subject)
 
     def is_none(self):
         """
         Ensures :attr:`subject` is ``None``.
         """
         self._run(unittest_case.assertIsNone, (self._subject,))
+        return ChainInspector(self._subject)
+
+    @property
+    def is_none_or(self):
+        """
+        Ensures :attr:`subject` is either ``None``, or satisfies subsequent (chained) conditions::
+
+            Ensure(None).is_none_or.is_an(int)
+        """
+        if self._subject is None:
+            return NoOpInspector(subject=self._subject, error_factory=self._error_factory)
+        else:
+            return self
 
     def is_not_none(self):
         """
         Ensures :attr:`subject` is not ``None``.
         """
         self._run(unittest_case.assertIsNotNone, (self._subject,))
+        return ChainInspector(self._subject)
 
     def is_empty(self):
         """
@@ -350,6 +395,7 @@ class Ensure(Inspector):
         """
         if len(self._subject) > 0:
             raise self._error_factory(_format("Expected {} to be empty", self._subject))
+        return ChainInspector(self._subject)
 
     def is_nonempty(self):
         """
@@ -357,6 +403,7 @@ class Ensure(Inspector):
         """
         if len(self._subject) == 0:
             raise self._error_factory(_format("Expected {} to be non-empty", self._subject))
+        return ChainInspector(self._subject)
 
     def is_a(self, prototype):
         """
@@ -365,6 +412,8 @@ class Ensure(Inspector):
         self._run(unittest_case.assertIsInstance, (self._subject, prototype))
         if hasattr(self._subject, '__iter__'):
             return IterableInspector(self._subject)
+        else:
+            return ReferenceInspector(self._subject)
 
     is_an = is_a
     is_an_instance_of = is_a
@@ -388,6 +437,7 @@ class Ensure(Inspector):
         Ensures :attr:`subject` is greater than 0.
         """
         self._run(unittest_case.assertGreater, (self._subject, 0))
+        return ChainInspector(self._subject)
 
     def is_a_positive(self, prototype):
         """
@@ -401,6 +451,7 @@ class Ensure(Inspector):
         Ensures :attr:`subject` is less than 0.
         """
         self._run(unittest_case.assertLess, (self._subject, 0))
+        return ChainInspector(self._subject)
 
     def is_a_negative(self, prototype):
         """
@@ -414,6 +465,7 @@ class Ensure(Inspector):
         Ensures :attr:`subject` is greater than or equal to 0.
         """
         self._run(unittest_case.assertGreaterEqual, (self._subject, 0))
+        return ChainInspector(self._subject)
 
     def is_a_nonnegative(self, prototype):
         """
@@ -427,6 +479,7 @@ class Ensure(Inspector):
         Ensures :attr:`subject` is not an object of class *prototype*.
         """
         self._run(unittest_case.assertNotIsInstance, (self._subject, prototype))
+        return ChainInspector(self._subject)
 
     not_a = is_not_a
 
@@ -436,6 +489,7 @@ class Ensure(Inspector):
         """
         if not re.match(pattern, self._subject, flags):
             raise self._error_factory(_format("Expected {} to match {}", self._subject, pattern))
+        return ChainInspector(self._subject)
 
     def is_an_iterable_of(self, prototype):
         """
@@ -478,8 +532,10 @@ class Ensure(Inspector):
         """
         Ensures :attr:`subject` is an int, float, or long.
         """
-        if not isinstance(self._subject, (int, float, long)):
-            raise self._error_factory(_format("Expected {} to be numeric (int, float, or long)", self._subject))
+        from decimal import Decimal
+        numeric_types = (int, float, long, Decimal) if USING_PYTHON2 else (int, float, Decimal)  # noqa
+        if not isinstance(self._subject, numeric_types):
+            raise self._error_factory(_format("Expected {} to be numeric (int, float, long or Decimal)", self._subject))
 
     def is_callable(self):
         """
@@ -496,6 +552,7 @@ class Ensure(Inspector):
             unittest_case.assertTrue(len(self._subject) == length)
         except self._catch as err:
             raise self._error_factory(_format("Expected {} to have length {}", self._subject, length))
+        return ChainInspector(self._subject)
 
     def is_greater_than(self, other):
         """
@@ -505,6 +562,7 @@ class Ensure(Inspector):
             unittest_case.assertTrue(self._subject > other)
         except self._catch as err:
             raise self._error_factory(_format("Expected {} to be greater than {}", self._subject, other))
+        return ChainInspector(self._subject)
 
     exceeds = is_greater_than
 
@@ -516,6 +574,7 @@ class Ensure(Inspector):
             unittest_case.assertTrue(self._subject < other)
         except self._catch as err:
             raise self._error_factory(_format("Expected {} to be less than {}", self._subject, other))
+        return ChainInspector(self._subject)
 
     def is_greater_than_or_equal_to(self, other):
         """
@@ -525,6 +584,7 @@ class Ensure(Inspector):
             unittest_case.assertTrue(self._subject >= other)
         except self._catch as err:
             raise self._error_factory(_format("Expected {} to be greater than or equal to {}", self._subject, other))
+        return ChainInspector(self._subject)
 
     def is_less_than_or_equal_to(self, other):
         """
@@ -534,6 +594,7 @@ class Ensure(Inspector):
             unittest_case.assertTrue(self._subject <= other)
         except self._catch as err:
             raise self._error_factory(_format("Expected {} to be less than or equal to {}", self._subject, other))
+        return ChainInspector(self._subject)
 
     def called_with(self, *args, **kwargs):
         """
@@ -559,20 +620,50 @@ class Ensure(Inspector):
         and the string representation of *expected_exception* must match regular expression *expected_regexp*.
         """
         return unittest_case.assertRaisesRegexp(expected_exception, expected_regexp, self._orig_subject,
-                                                  *self._args, **self._kwargs)
+                                                *self._args, **self._kwargs)
 
-    #def has_schema(self, schema):
+    def is_a_numeric_string(self):
+        return self.is_a(NumericString)
+
+    def is_a_numeric_bytestring(self):
+        return self.is_a(NumericByteString)
+
+    def is_an_integer_string(self):
+        return self.is_an(IntegerString)
+
+    def is_an_integer_bytestring(self):
+        return self.is_an(IntegerByteString)
+
+    # def has_schema(self, schema):
     #    import jsonschema
     #    self._run(jsonschema.validate, (self._subject, schema))
 
+    def satisfies(self, predicate, *args, **kwargs):
+        def run():
+            if isinstance(predicate, string_types):
+                # `predicate` can be a string that names a method.
+                # In this case, look up the method and use that as the actual predicate.
+                if not predicate.startswith('.'):
+                    raise TypeError('Predicate must be a callable or method name, '
+                                    'and method names must start with ".": {}'.format(predicate))
+                method_name = predicate[1:]
+                def actual_predicate(subj, *args):
+                    return getattr(subj, method_name)(*args, **kwargs)
+                predicate_name = predicate
+            else:
+                # Otherwise assume `predicate` is some kind of callable
+                actual_predicate = predicate
+                predicate_name = getattr(predicate, '__name__', predicate)
 
-class Check(object):
-    """
-    Like Ensure, but if a check fails, saves the error instead of raising it immediately. The error can then be acted
-    upon using :meth:`or_raise()` or :meth:`or_call()`.
+            if not actual_predicate(self._subject, *args, **kwargs):
+                # Note the error message uses the original `predicate`
+                msg = _format("Expected {} to satisfy {}", self._subject, predicate_name)
+                if args:
+                    msg += _format(" on args {}", args)
+                raise ValueError(msg)
+        return self._run(run)
 
-    ``.each_of()`` is not supported by the **Check** inspector; all other methods are supported.
-    """
+class InspectorProxy(object):
     def __init__(self, *args, **kwargs):
         self._inspector = Ensure(*args, **kwargs)
         self._exception = None
@@ -583,7 +674,7 @@ class Check(object):
 
     def __getattr__(self, item):
         inspect_method = getattr(self._inspector, item)
-        if isinstance(inspect_method, Inspector):
+        if isinstance(inspect_method, (Inspector, NoOpInspector)):
             self._inspector = inspect_method
             return self
 
@@ -596,20 +687,32 @@ class Check(object):
             return self
         return inspect
 
+class Check(InspectorProxy):
+    """
+    Like Ensure, but if a check fails, saves the error instead of raising it immediately. The error can then be acted
+    upon using :meth:`or_raise()` or :meth:`or_call()`.
+
+    ``.each_of()`` is not supported by the **Check** inspector; all other methods are supported.
+    """
     def or_raise(self, error_factory, message=None, *args, **kwargs):
         """
         Raises an exception produced by **error_factory** if a predicate fails.
 
-        :param error_factory: Class or callable (e.g. :class:`Exception`) which will be invoked to produce the resulting exception.
-        :param message: String to be formatted and passed as the first argument to *error_factory*. The following substrings are replaced by formatting: "{type}" by the exception type, "{msg}" by the exception's string representation, and "{}" by both. If *message* is ``None``, the original exception is passed.
+        :param error_factory:
+            Class or callable (e.g. :class:`Exception`) which will be invoked to produce the resulting exception.
+            You can define a custom callable here; it will be given the underlying predicate's exception (AssertError)
+            as the first argument, followed by any arguments passed to ``or_raise``.
+        :param message:
+            String to be formatted and passed as the first argument to *error_factory*. If this is given, subsequent
+            arguments passed to ``or_raise`` will be used to format contents of the string, and will not be passed to
+            ``error_factory``. The keyword argument ``error`` will be set to the underlying predicate's exception.
         """
         if self._exception:
             if message is None:
                 raise error_factory(self._exception, *args, **kwargs)
             else:
-                raise error_factory(message.format('{}: {}'.format(type(self._exception).__name__, self._exception),
-                                                   msg=self._exception, type=type(self._exception).__name__),
-                                    *args, **kwargs)
+                kwargs.setdefault("error", self._exception)
+                raise error_factory(message.format(*args, **kwargs))
 
     otherwise = or_raise
 
@@ -620,6 +723,11 @@ class Check(object):
         if self._exception:
             _callable(*args, **kwargs)
 
+class NoOpInspector(InspectorProxy):
+    """
+    Goes through the motions of Ensure, but never raises on the error condition. For chaining with conditional
+    predicates like ``is_none_or``.
+    """
 
 def _check_default_argument(f, arg, value):
     if value is not None and arg in f.__annotations__:
@@ -691,8 +799,8 @@ class WrappedFunctionReturn(WrappedFunction):
 
         return_val = self.f(*args, **kwargs)
         if not isinstance(return_val, self.return_templ):
-                msg = "Return value of {f} does not match annotation type {t}"
-                raise EnsureError(msg.format(f=self.f, t=self.return_templ))
+            msg = "Return value of {f} does not match annotation type {t}"
+            raise EnsureError(msg.format(f=self.f, t=self.return_templ))
         return return_val
 
 
