@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # coding: utf-8
 
-from __future__ import print_function, unicode_literals
+from __future__ import absolute_import, division, print_function, unicode_literals
 
 import os
 import sys
@@ -9,9 +9,10 @@ import unittest
 import collections
 import copy
 import re
+from six import text_type as str
 
 sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
-from ensure import *
+from ensure import ensure, check, Ensure, Check, EnsureError, ensure_annotations
 
 
 class TestEnsure(unittest.TestCase):
@@ -39,8 +40,8 @@ class TestEnsure(unittest.TestCase):
                       'Ensure(x).does_not_contain(-1)',
                       'Ensure(x).does_not_equal(range(10))',
                       'Ensure(x).has_attribute("__iter__")',
-                      'Ensure(x).has_length(10)',
-                      'Ensure(x).is_nonempty()',
+                      'Ensure(x).has_length(10).also.is_nonempty()',
+                      'Ensure(x).is_nonempty().also.has_length(10)',
                       'Ensure(x).is_a(collections.Mapping)',
                       'Ensure(x).is_a_dict_of(int).to(int)',
                       'Ensure(x).is_a(collections.Mapping).of(int).to(int)',
@@ -55,6 +56,7 @@ class TestEnsure(unittest.TestCase):
                       'Ensure(-1).is_a_negative(int)',
                       'Ensure(0).is_nonnegative()',
                       'Ensure(0).is_a_nonnegative(int)',
+                      'Ensure(1).is_a_positive(int).which.equals(1.0)',
                       'Ensure((collections.namedtuple("Thing", ["x"]))(x={})).has_attribute("x").which.is_a(dict)',
                       'Ensure({1:"a"}).has_key(1).whose_value.has_length(1)',
                       'Ensure({1: "a", 2: "b", 3: "c"}).has_keys((1, 2))',
@@ -63,16 +65,27 @@ class TestEnsure(unittest.TestCase):
                       'Ensure(os.path.join).called_with("a", "b").returns(os.path.join("a", "b"))',
                       'Ensure(int).called_with("1100101", base=2).returns(101)',
                       'Ensure.each_of([1,2,3]).is_an(int)',
-                      'Ensure.each_of([lambda x: x, lambda y: y]).called_with(1).returns(1)')
+                      'Ensure.each_of([lambda x: x, lambda y: y]).called_with(1).returns(1)',
+                      'Ensure(True).is_none_or.is_an(int)',  # See https://www.python.org/dev/peps/pep-0285/ (section 6)
+                      'Ensure(None).is_none_or.is_a_negative(int)',
+                      'Ensure(-5).is_none_or.is_a_negative(int)',
+                      'Ensure({"a": "b"}).is_none_or.has_key("a")',
+                      'Ensure("A").satisfies(str.isupper)',
+                      'Ensure("A").satisfies(".isupper")',
+                      'Ensure("ABC").satisfies(str.startswith, "AB")',
+                      'Ensure("ABC").satisfies(".startswith", "AB")',
+                      'Ensure(3).satisfies(lambda x, y: x < y, y=4)')
 
         for clause in ok_clauses:
             print("Testing OK clause", clause)
             eval(clause)
             if 'each_of' not in clause:
                 for sub in r'Check\1.otherwise(Exception)', r'Check\1.or_raise(Exception)', r'Check\1.or_call(self.assertTrue, False)':
+                    print("Testing OK clause", re.sub(r'^Ensure(.+)', sub, clause))
                     eval(re.sub(r'^Ensure(.+)', sub, clause))
 
         bad_clauses = ('Ensure(x).contains(-1)',
+                       'Ensure(x).has_length(10).also.is_empty()',
                        'Ensure(x).contains_all_of(range(20))',
                        'Ensure(x).contains_no(int)',
                        'Ensure(x).contains_none_of(range(0, 30))',
@@ -104,7 +117,16 @@ class TestEnsure(unittest.TestCase):
                        'Ensure({1: "a", 2: "b"}).has_only_keys([1, 2, 3])',
                        'Ensure([1, 2, 3]).has_only_keys([1, 2, 3])',
                        'Ensure(os.path.join).called_with("a", "b").returns(None)',
-                       'Ensure.each_of([lambda x: x, lambda y: y]).called_with(2).returns(1)')
+                       'Ensure(1).is_a_positive(int).which.equals(1.2)',
+                       'Ensure.each_of([lambda x: x, lambda y: y]).called_with(2).returns(1)',
+                       'Ensure(5).is_none_or.is_a_negative(int)',
+                       'Ensure(None).is_a_negative(int)',
+                       'Ensure("a").satisfies(str.isupper)',
+                       'Ensure("a").satisfies(".isupper")',
+                       'Ensure("ABC").satisfies(str.startswith, "Z")',
+                       'Ensure("ABC").satisfies(".startswith", "Z")',
+                       'Ensure(5).satisfies(str.isupper)',
+                       'Ensure(5).satisfies(".isupper")')
 
         for clause in bad_clauses:
             print("Testing bad clause", clause)
@@ -113,6 +135,7 @@ class TestEnsure(unittest.TestCase):
             if 'each_of' not in clause:
                 for sub in r'Check\1.otherwise(Exception)', r'Check\1.or_raise(Exception)', r'Check\1.or_call(self.assertTrue, False)':
                     with self.assertRaises(Exception):
+                        print("Testing bad clause", re.sub(r'^Ensure(.+)', sub, clause))
                         eval(re.sub(r'^Ensure(.+)', sub, clause))
 
         with self.assertRaises(EnsureError):
@@ -123,6 +146,37 @@ class TestEnsure(unittest.TestCase):
             Ensure(x).called_with().is_an(int)
         Ensure(lambda: True).is_callable()
 
+        Ensure("1.1").is_a_numeric_string()
+        with self.assertRaises(EnsureError):
+            Ensure(b"1").is_a_numeric_string()
+        with self.assertRaises(EnsureError):
+            Ensure("").is_a_numeric_string()
+        with self.assertRaises(EnsureError):
+            Ensure(None).is_a_numeric_string()
+
+        Ensure(b"1").is_a_numeric_bytestring()
+        Ensure(b"1.1").is_a_numeric_bytestring()
+        with self.assertRaises(EnsureError):
+            Ensure("1").is_a_numeric_bytestring()
+        with self.assertRaises(EnsureError):
+            Ensure(b"").is_a_numeric_bytestring()
+        with self.assertRaises(EnsureError):
+            Ensure(None).is_a_numeric_bytestring()
+
+        Ensure("1").is_an_integer_string()
+        with self.assertRaises(EnsureError):
+            Ensure("1.1").is_an_integer_string()
+
+        Ensure(b"1").is_an_integer_bytestring()
+        with self.assertRaises(EnsureError):
+            Ensure(b"1.1").is_an_integer_bytestring()
+        with self.assertRaises(EnsureError):
+            Ensure("1").is_an_integer_bytestring()
+        with self.assertRaises(EnsureError):
+            Ensure(b"").is_an_integer_bytestring()
+        with self.assertRaises(EnsureError):
+            Ensure(None).is_an_integer_bytestring()
+
     def test_called_with(self):
         for i in None, True, 1, {}, [], lambda: True:
             with self.assertRaises(EnsureError):
@@ -301,6 +355,9 @@ class C(object):
         with self.assertRaisesRegex(EnsureError, "Argument x to <function (C.g|g) at .+> does not match annotation type <class 'int'>"):
             g = C().g(3.2, 1)
 
+    def test_error_formatting(self):
+        with self.assertRaisesRegexp(Exception, "Major fail detected"):
+            check(False).is_true().or_raise(KeyError, "{} {error} detected", "Major", error="fail")
 
 if __name__ == '__main__':
     unittest.main()
